import os
import time
import toml
import socket
import xml.etree.ElementTree as ET
from typing import Any

import rclpy
import roboticstoolbox as rtb
from rclpy.node import Node
from sensor_msgs.msg import JointState


class JointNameListener(Node):
    joint_names: list[str]|None
    def __init__(self):
        super().__init__('joint_name_listener')
        self.subscription = self.create_subscription(
            JointState,
            '/joint_states',
            self.joint_state_callback,
            1
        )
        self.joint_names = None

    def joint_state_callback(self, msg: JointState):
        self.joint_names = list(msg.name)

def interactive_input() -> tuple[dict[str, Any], rtb.robot]:
    config = {}
    rconf = {}
    nconf = {}
    robot = None
    while True:
        use_urdf = input("Do you have a URDF file you want to use? (y/n): ").strip().lower()
        if use_urdf == 'y':
            while True:
                rconf["urdf"] = input("Enter the absolute path to the URDF file: ")
                if os.path.isfile(rconf["urdf"]):
                    if not rconf["urdf"].endswith('.urdf'):
                        print("The file is not a URDF file. Please enter a valid URDF file.")
                        continue
                    break
                else:
                    print("Invalid path. Please enter a valid path to the URDF file.")
            tree = ET.parse(rconf["urdf"])
            root = tree.getroot()
            robot = rtb.ERobot.URDF(rconf["urdf"])
            rconf["joint_names"] = [joint.get('name') for joint in root.findall('joint') if joint.get('type') == 'revolute' or joint.get('type') == 'continuous' or joint.get('type') == 'prismatic']
            print(robot)
            rconf["joint_velocity_limits"] = {}
            # Iterate over all joints in the URDF
            for joint in root.findall('.//joint'):
                joint_name = joint.get('name')  # Get the name of the joint
                
                # Look for the <limit> tag under each joint
                limit = joint.find('limit')
                
                if limit is not None:
                    # Extract the velocity limit (if it exists)
                    velocity_limit = limit.get('velocity')
                    
                    if velocity_limit is not None:
                        rconf["joint_velocity_limits"][joint_name] = float(velocity_limit)
            
            while True:
                try:
                    print('-+'*50)
                    print("The cost mask determines which coordinates are used for the IK.\nEach element of the cost mask corresponds to a Cartesian coordinate [x, y, z, roll, pitch, yaw].")
                    print("The cost mask 111000 means that the IK will only consider translation and no rotaion.")
                    cost_mask_string = input(f"Enter the cost mask (6 integers (1 or 0), of which <= {robot.n} are 1): ")
                    rconf["cost_mask"] = [int(i) for i in cost_mask_string]
                    if sum(rconf["cost_mask"]) <= robot.n and len(rconf["cost_mask"]) == 6:
                        break
                    else:
                        print(f"Invalid input. Expected 6 integers of which {robot.n if robot.n < 6 else 6} or less are 1.")
                except ValueError:
                    print("Invalid input. Please enter integers only.")
            print(f"The following coordinates will be used for the IK: {[j for i,j in enumerate(['x','y','z','roll','pitch','yaw']) if rconf['cost_mask'][i]==1]}")
            break
        elif use_urdf == 'n':
            node = JointNameListener()
            print("Wainting 10 sec for JointState messages to extract joint names...")
            rclpy.spin_once(node)
            counter = 0
            while not(node.joint_names):
                if counter > 100:
                    rconf["joint_names"] = None
                    break
                counter+=1
                time.sleep(0.1)
                rclpy.spin_once(node)
            rconf["joint_names"] = node.joint_names
            node.destroy_node()
            if rconf["joint_names"]:
                correct_names = input("The following joint names were found:\n" + "\n".join(rconf["joint_names"]) + "\nAre these correct? (y/n): ").strip().lower()
                while True:
                    if correct_names == 'y':
                        break
                    elif correct_names == 'n':
                        rconf["joint_names"] = None
                        break
                    correct_names = input("Invalid input. Please enter 'y' or 'n'.")
            if not(rconf["joint_names"]):
                print("Please enter the joint names manually.")
                while True:
                        rconf["joint_names"] = []
                        print('-+'*50)
                        print("Enter the joint names manually one by one. Type 'done' when you are finished:")
                        print("Attention: The order of the joints is important. It should be the same in which you want to send the joint positions.")
                        while True:
                            print('-'*50)
                            joint_name = input("Enter joint name (or 'done' to finish): ").strip()
                            if joint_name.lower() == 'done':
                                break
                            if joint_name:
                                rconf["joint_names"].append(joint_name)
                        print('-+'*50)
                        correct_names = input(f"Are those the joint names as defined in your JointTrajectroy recieving Node?: {rconf['joint_names']}. (y/n)?: ").strip()
                        if correct_names.lower() == 'y':
                            break
                        else:
                            print("Please re-enter the joint names.")
            while True:
                try:
                    rconf["joint_velocity_limits"] = {}
                    for name in rconf["joint_names"]:
                        while True:
                            try:
                                print('--'*50)
                                limit = input(f"Enter the velocity limit for joint '{name}': ").strip()
                                if limit == "":
                                    continue
                                else:
                                    rconf["joint_velocity_limits"][name] = float(limit)
                                    break
                            except ValueError:
                                print("Invalid input. Please enter a numeric value or press Enter to skip.")
                    break
                except ValueError:
                    print("Invalid input. Please enter numeric values or leave blank to skip.")
            robot = None
            rconf["cost_mask"] = None
            break
        print("Invalid input. Please enter 'y' or 'n'.")
    if robot:
        while True:
            print('+-' * 50)
            set_limits = input("Do you want to set limits for x, y and z? (y/n): ").strip().lower()
            if set_limits == 'y':
                while True:
                    try:
                        rconf["x_limits"] = [float(i) if i != 'x' else None for i in input("Enter the lower and upper limits for x (space-separated, enter 'x' for no limit): ").split()]
                        rconf["y_limits"] = [float(i) if i != 'x' else None for i in input("Enter the lower and upper limits for y (space-separated, enter 'x' for no limit): ").split()]
                        rconf["z_limits"] = [float(i) if i != 'x' else None for i in input("Enter the lower and upper limits for z (space-separated, enter 'x' for no limit): ").split()]
                    
                        if len(rconf["x_limits"]) != 2 or len(rconf["y_limits"]) != 2 or len(rconf["z_limits"]) != 2:
                            print("Invalid input. Please enter exactly two values (or leave blank) for each limit.")
                            continue
                    
                        if (rconf["x_limits"][0] is not None and rconf["x_limits"][1] is not None and rconf["x_limits"][0] >= rconf["x_limits"][1]) or \
                        (rconf["y_limits"][0] is not None and rconf["y_limits"][1] is not None and rconf["y_limits"][0] >= rconf["y_limits"][1]) or \
                        (rconf["z_limits"][0] is not None and rconf["z_limits"][1] is not None and rconf["z_limits"][0] >= rconf["z_limits"][1]):
                            print("Invalid input. Lower limit must be less than upper limit for each axis.")
                            continue
                    
                        print("Current limits:")
                        print(f"x: {rconf['x_limits']}")
                        print(f"y: {rconf['y_limits']}")
                        print(f"z: {rconf['z_limits']}")
                        con = True
                        while con:
                            confirm = input("Do you want your TCP to move in this range? (y/n): ").strip().lower()
                            if confirm == 'y':
                                break
                            elif confirm == 'n':
                                print("Please re-enter the limits.")
                                con = False
                            else:
                                print("Invalid input. Please enter 'y' or 'n'.")
                        if con:
                            break
                    except ValueError:
                        print("Invalid input. Please enter numeric values only.")
                    break
            elif set_limits == 'n': 
                rconf["x_limits"] = [None, None]
                rconf["y_limits"] = [None, None]
                rconf["z_limits"] = [None, None]
                break
            print("Invalid input. Please enter 'y' or 'n'.")

        while True:
            print('+-' * 50)
            set_limits = input("Do you want to set workspace limits in x, y and z direction? (y/n): ").strip().lower()
            if set_limits == 'y':
                while True:
                    try:
                        rconf["x_limits_workspace"] = [float(i) if i != 'x' else None for i in input("Enter the lower and upper limits for x (space-separated, enter 'x' for no limit): ").split()]
                        rconf["y_limits_workspace"] = [float(i) if i != 'x' else None for i in input("Enter the lower and upper limits for y (space-separated, enter 'x' for no limit): ").split()]
                        rconf["z_limits_workspace"] = [float(i) if i != 'x' else None for i in input("Enter the lower and upper limits for z (space-separated, enter 'x' for no limit): ").split()]
                    
                        if len(rconf["x_limits_workspace"]) != 2 or len(rconf["y_limits_workspace"]) != 2 or len(rconf["z_limits_workspace"]) != 2:
                            print("Invalid input. Please enter exactly two values (or leave blank) for each limit.")
                            continue
                    
                        if (rconf["x_limits_workspace"][0] is not None and rconf["x_limits_workspace"][1] is not None and rconf["x_limits_workspace"][0] >= rconf["x_limits_workspace"][1]) or \
                        (rconf["y_limits_workspace"][0] is not None and rconf["y_limits_workspace"][1] is not None and rconf["y_limits_workspace"][0] >= rconf["y_limits_workspace"][1]) or \
                        (rconf["z_limits_workspace"][0] is not None and rconf["z_limits_workspace"][1] is not None and rconf["z_limits_workspace"][0] >= rconf["z_limits_workspace"][1]):
                            print("Invalid input. Lower limit must be less than upper limit for each axis.")
                            continue
                    
                        print("Current limits:")
                        print(f"x: {rconf['x_limits_workspace']}")
                        print(f"y: {rconf['y_limits_workspace']}")
                        print(f"z: {rconf['z_limits_workspace']}")
                        con = True
                        while con:
                            confirm = input("Do you want your robot to move in this range? (y/n): ").strip().lower()
                            if confirm == 'y':
                                break
                            elif confirm == 'n':
                                print("Please re-enter the limits.")
                                con = False
                            else:
                                print("Invalid input. Please enter 'y' or 'n'.")
                        if con:
                            break
                    except ValueError:
                        print("Invalid input. Please enter numeric values only.")
                break
            elif set_limits == 'n': 
                rconf["x_limits_workspace"] = [None, None]
                rconf["y_limits_workspace"] = [None, None]
                rconf["z_limits_workspace"] = [None, None]
                break
            print("Invalid input. Please enter 'y' or 'n'.")
                    # Ask the user if they want to set new joint limits
        while True:
            print('+-'*50)
            update_limits = input("Do you want to set new joint limits? (y/n): ").strip().lower()
            if update_limits == 'y':
                for i in range(len(rconf["joint_names"])):
                    while True:
                        try:
                            lim = robot.qlim.copy()
                            # Find the link corresponding to the joint name
                            print("-" * 50)
                            print(f"Current position limits for joint '{rconf['joint_names'][i]}': [{robot.qlim[0][i]} {robot.qlim[1][i]}] rad")
                            lower_limit = input(f"Enter the new lower limit for joint '{rconf['joint_names'][i]}' (or press Enter to keep current): ").strip()
                            upper_limit = input(f"Enter the new upper limit for joint '{rconf['joint_names'][i]}' (or press Enter to keep current): ").strip()
                            
                            if lower_limit and upper_limit and float(lower_limit) >= float(upper_limit):
                                print("Invalid input. Lower limit must be less than upper limit.")
                                continue
                            lower_limit = float(lower_limit) if lower_limit!="" else None
                            upper_limit = float(upper_limit) if upper_limit!="" else None
                            if lower_limit is not None:
                                if lower_limit<lim[0][i]:
                                    while True:
                                        sure = input(f"Are you sure you want to set the lower limit to {lower_limit} rad which is less than the default limit {lim[0][i]} (y/n)?: ").strip().lower()
                                        if sure == 'y':
                                            lim[0][i] = float(lower_limit)
                                            break
                                        elif sure == 'n':
                                            print("Lower limit not changed.")
                                            break
                                        print("Invalid input. Please enter 'y' or 'n'.")
                                else:
                                    lim[0][i] = float(lower_limit)
                            if upper_limit is not None:
                                if upper_limit>lim[1][i]:
                                    while True:
                                        sure = input(f"Are you sure you want to set the upper limit to {upper_limit} rad which is more than the default limit {lim[1][i]} (y/n)?: ").strip().lower()
                                        if sure == 'y':
                                            lim[1][i] = float(upper_limit)
                                            break
                                        elif sure == 'n':
                                            print("Upper limit not changed.")
                                            break
                                        print("Invalid input. Please enter 'y' or 'n'.")
                                else:
                                    lim[1][i] = float(upper_limit)
                            robot.qlim = lim
                            print(f"New limits for joint '{rconf['joint_names'][i]}': [{robot.qlim[0][i]} {robot.qlim[1][i]}] rad")
                            print("-" * 50)
                            break
                        except ValueError:
                            print("Invalid input. Please enter numeric values or leave blank to keep current limits.")
                break
            if update_limits == 'n':
                break
            print("Invalid input. Please enter 'y' or 'n'.")
        '''
        use_link_mask = input("Do you want to use a link mask? (y/n): ").strip().lower()
        if use_link_mask == 'y':
            while True:
                try:
        '''
    else:
        while True:
            print('+-' * 50)
            update_limits = input("Do you want to set joint limits? (y/n): ").strip().lower()
            n_joints = len(rconf["joint_names"])
            if update_limits == 'y':
                rconf["joint_lim"] = [[]*n_joints,[]*n_joints]
                for i, joint in enumerate(rconf["joint_names"]):
                    while True:
                        try:
                            print("-" * 50)
                            lower_limit = input(f"Enter the new lower limit for joint '{joint}' (or press Enter for None): ").strip()
                            upper_limit = input(f"Enter the new upper limit for joint '{joint}' (or press Enter for None): ").strip()
                            
                            if lower_limit and upper_limit and float(lower_limit) >= float(upper_limit):
                                print('--' * 50)
                                print("      !!!      Invalid input. Lower limit must be less than upper limit.      !!!      ")
                                continue
                            rconf["joint_lim"][0][i] = float(lower_limit) if lower_limit!="" else None
                            rconf["joint_lim"][1][i] = float(upper_limit) if upper_limit!="" else None
                            break
                        except ValueError:
                            print("Invalid input. Please enter numeric values or leave blank to keep current limits.")
                print(f'New limits for joint:\n lower: {rconf["joint_lim"][0]}\n upper: {rconf["joint_lim"][1]}')
                break
            elif update_limits == 'n':
                rconf["joint_lim"] = None
                break
            print("Invalid input. Please enter 'y' or 'n'.")
    while True:
        try:
            print('+-' * 50)
            rconf["trajectory_topic_name"] = input("Enter the topic name to which the joint trajectory should be sent (press Enter for default: '/scaled_joint_trajectory_controller/joint_trajectory'): ").strip()
            if rconf["trajectory_topic_name"] == "":
                rconf["trajectory_topic_name"] = '/scaled_joint_trajectory_controller/joint_trajectory'
                break
            elif rconf["trajectory_topic_name"].startswith("/"):
                break
            else:
                print("Invalid topic name. A valid topic name should start with '/'.")
        except Exception as e:
            print(f"An error occurred: {e}")
        print('--' * 50)
    while True:
        try:
            print('+-' * 50)
            rconf["hz"] = input("Enter the desired refresh frequency (Hz) (or press Enter for default: 100): ")
            if rconf["hz"] == "":
                rconf["hz"] = 100
            else:
                rconf["hz"] = float(rconf["hz"])
            break
        except ValueError:
            print("Invalid input. Please enter a valid number.")
            continue
    while True:
        try:
            print('+-' * 50)
            nconf["log_ip"] = str(input("Enter the target IP on which you want to recieve the log as OSC Messages (or press Enter for default: '127.0.0.1'): "))
            if nconf["log_ip"] == "":
                nconf["log_ip"] = "127.0.0.1"
            print('--' * 50)
            nconf["log_port"] = input("Enter the target port for the log messages (or press Enter for default: 5005): ")
            if nconf["log_port"] == "":
                nconf["log_port"] = 5005
            else:
                nconf["log_port"] = int(nconf["log_port"])
            break
        except ValueError:
            print("Invalid input. Please enter a valid IP address.")
            continue
    while True:
        try:
            print('+-' * 50)
            nconf["state_ip"] = str(input("Enter the target IP on which you want to recieve the joint states as OSC Messages (or press Enter for default: '127.0.0.1'): "))
            if nconf["state_ip"] == "":
                nconf["state_ip"] = "127.0.0.1"
            print('--' * 50)
            nconf["state_port"] = input("Enter the target port for the joint state messages (or press Enter for default: 7000): ")
            if nconf["state_port"] == "":
                nconf["state_port"] = 7000
            else:
                nconf["state_port"] = int(nconf["state_port"])
            break
        except ValueError:
            print("Invalid input. Please enter a valid IP address.")
            continue
    while True:
        try:
            print('+-' * 50)
            nconf["commands_port"] = input("Enter the port you want to send your commands to (or press Enter for default: 8000): ")
            if nconf["commands_port"] == nconf["state_port"]:
                print("The commands port must be different from the state port.")
                continue
            if nconf["commands_port"] == "":
                nconf["commands_port"] = 8000
            else:
                nconf["commands_port"] = int(nconf["commands_port"])
            break
        except ValueError:
            print("Invalid input. Please enter a valid port.")
            continue
    print()
    print('=-=' * 50)
    print()
    print(f"Sending joint state's' to {nconf['state_ip']}:{nconf['state_port']}")
    print()
    print('=-=' * 50)
    print()
    print(f"Sending log messages to {nconf['log_ip']}:{nconf['log_port']}")
    print()
    print('=-=' * 50)
    print()
    print(f"Ready to receive OSC messages on {socket.gethostbyname(socket.gethostname())}:{nconf['commands_port']}")
    print()
    print('=-=' * 50)
    print()
    config["robot"] = rconf
    config["network"] = nconf
    while True:
        save_config = input("Do you want to save this interface? (y/n): ").strip().lower()
        if save_config == "y":
            config_path = input("Where do you want to save the config?: ").strip()
            with open(f"{config_path}.toml", "w") as f:
                toml.dump(config, f)
            break
        elif save_config == "n":
            print("Will not be saving config")
        print("Invalid input. Please enter 'y' or 'n'.")
    return config, robot
