#ruff: noqa: F403, F405
import rclpy
import argparse
import toml
from rclpy.node import Node
from trajectory_msgs.msg import JointTrajectory, JointTrajectoryPoint
from sensor_msgs.msg import JointState
from rcl_interfaces.msg import Log
from osc4py3.as_allthreads import *
from osc4py3 import oscmethod as osm
import numpy as np
import spatialmath as sm
import roboticstoolbox as rtb
from osc4py3 import oscbuildparse
from osc_ros2.interactive_input import interactive_input
import time
import re
import socket


class OSC_ROS2_interface(Node):
    """Node to publish joint trajectories based on OSC messages."""
    
    def __init__(self, robot, config):
        super().__init__('scaled_joint_trajectory_publisher')
        self.subscription = self.create_subscription(
            JointState,
            '/joint_states',
            self.joint_states_callback,
            1
        )
        self.subscription = self.create_subscription(
            Log,
            '/rosout',
            self.log_callback,
            100
        )
        # Store received joint positions
        self.current_joint_positions = None
        self.joint_names = config["robot"]["joint_names"]
        self.joint_velocity_limits = config["robot"]["joint_velocity_limits"]
        self.cost_mask = config["robot"]["cost_mask"]
        self.trajectory_topic_name = config["robot"]["trajectory_topic_name"]
        self.hz = config["robot"]["hz"]
        self.state_ip = config["network"]["state_ip"]
        self.state_port = config["network"]["state_port"]
        self.log_ip = config["network"]["log_ip"]
        self.log_port = config["network"]["log_port"]
        self.commands_port = config["network"]["commands_port"]
        float_or_none = lambda v: None if "None" in v else float(v)
        self.x_limits = list(map(float_or_none, config["robot"]["x_limits"]))
        self.y_limits = list(map(float_or_none, config["robot"]["y_limits"]))
        self.z_limits = list(map(float_or_none, config["robot"]["z_limits"]))
        self.x_limits_workspace = list(map(float_or_none, config["robot"]["x_limits_workspace"]))
        self.y_limits_workspace = list(map(float_or_none, config["robot"]["y_limits_workspace"]))
        self.z_limits_workspace = list(map(float_or_none, config["robot"]["z_limits_workspace"]))
        self.robot = robot
        self.desired = None
        self.previous_desired = None
        self.log_dict = {
            10: "DEBUG",
            20: "INFO",
            30: "WARN",
            40: "ERROR",
            50: "FATAL",
        }
        self.speed_scaling = 0.2
        self.new = False
        self.n_joints = len(self.joint_names)   
        # ROS2 Publisher
        self.publisher = self.create_publisher(
            JointTrajectory, 
            self.trajectory_topic_name, 
            1
        )
        osc_startup()
        osc_udp_client(self.state_ip, self.state_port, "osc_client")
        osc_udp_client(self.log_ip, self.log_port, "osc_log_client")
        osc_udp_server('0.0.0.0', self.commands_port, "osc_server")
        # Register OSC handler
        osc_method("/joint_positions", self.joint_positions_handler, argscheme=osm.OSCARG_DATAUNPACK)
        osc_method("/joint_position/*", self.joint_position_handler, argscheme=osm.OSCARG_ADDRESS+osm.OSCARG_DATAUNPACK)
        osc_method("/tcp_coordinates", self.tcp_coordinates_handler, argscheme=osm.OSCARG_DATAUNPACK)
        osc_method("/joint_trajectory", self.joint_trajectory_handler, argscheme=osm.OSCARG_DATAUNPACK)
        osc_method("/cartesian_trajectory", self.cartesian_trajectory_handler, argscheme=osm.OSCARG_DATAUNPACK)
        osc_method("/speed_scaling", self.speed_scaling_handler, argscheme=osm.OSCARG_DATAUNPACK)
        self.get_logger().info(f"Publishing joint trajectory to {self.trajectory_topic_name}")
        self.get_logger().info(f'Ready to receive OSC messages on {socket.gethostbyname(socket.gethostname())}:{self.commands_port}')
        self.get_logger().info(f'Sending joint states to {self.state_ip}:{self.state_port}')
        self.get_logger().info(f'Sending log messages to {self.log_ip}:{self.log_port}')
        self.create_timer(1/self.hz, self.update_position)  # Timer to update the position
        self.create_timer(3, self.reset_prev)  # reset the previous desired position

    def reset_prev(self): self.previous_desired = None

    def speed_scaling_handler(self, *args):
        """Handles incoming OSC messages for speed scaling."""
        try:
            if len(args) == 1:
                if args[0] < 0:
                    self.speed_scaling = -float(args[0])
                else:
                    self.speed_scaling = float(args[0])
                if self.speed_scaling > 1:
                    self.get_logger().warn(f"speed_scaling_handler: Attention! Speed scaling {self.speed_scaling} is greater than 1!")
                self.get_logger().info(f"Speed scaling set to {self.speed_scaling}")
            else:
                self.get_logger().warn(f"Invalid number of arguments for speed scaling. Expected 1, but got {len(args)}.")
        except Exception as e:
            self.get_logger().fatal(f"speed_scaling_handler: {e}")
    
    def joint_trajectory_handler(self, *args):
        try:
            if len(args[0]) == 6:
                points = [[float(j) for j in i] for i in args]
            elif len(args[0]) >= 7:
                points = [[float(j) for j in i[:6]] for i in args]
                self.get_logger().warn("joint_trajectory_handler: Duration is not supported for joint trajectory yet. Ignoring duration.")
            else:
                self.get_logger().warn(f"joint_trajectory_handler: Invalid number of arguments for joint trajectory. Expected {self.n_joints} ([q0, q1, q2, ..., q{self.n_joints}]) or {self.n_joints+1} ([q0, q1, q2, ..., q{self.n_joints}, duration]), but got {len(args[0])}.")
                return
            self.desired = ["joint_trajectory"] + points
            self.new = True
        except Exception as e:
            self.get_logger().fatal(f"joint_trajectory_handler: {e}")

    def joint_position_handler(self, address, *args):
        """Handles incoming OSC messages for joint positions."""
        try:
            joint_name = address.split("/")[-1]
            if joint_name in self.joint_names:
                if len(args) == 1:
                    position = float(args[0])
                    index = self.joint_names.index(joint_name)
                    if self.robot:
                        if position < self.robot.qlim[0][index]:
                            position = self.robot.qlim[0][index]
                            self.get_logger().warn(
                                f"joint_position_handler: Joint '{joint_name}' position {position} is out of bounds. Using {self.robot.qlim[0][index]}."
                            )
                        elif position > self.robot.qlim[1][index]:
                            position = self.robot.qlim[1][index]
                            self.get_logger().warn(
                                f"joint_position_handler: Joint '{joint_name}' position {position} is out of bounds. Using {self.robot.qlim[1][index]}."
                            )
                    else:
                        if self.joint_lim[0][index] is not None and position < self.joint_lim[0][index]:
                            position = self.joint_lim[0][index]
                            self.get_logger().warn(
                                f"joint_position_handler: Joint '{joint_name}' position {position} is out of bounds. Using {self.joint_lim[0][index]}."
                            )
                        elif self.joint_lim[1][index] is not None and position > self.joint_lim[1][index]:
                            position = self.joint_lim[1][index]
                            self.get_logger().warn(
                                f"joint_position_handler: Joint '{joint_name}' position {position} is out of bounds. Using {self.joint_lim[1][index]}."
                            )
                    desired_joint_positions = self.current_joint_positions
                    desired_joint_positions[index] = position
                    self.desired = ["joint_positions"] + desired_joint_positions
                    self.new = True
                elif len(args) == 2:
                    position = float(args[0])
                    duration = float(args[1])
                    index = self.joint_names.index(joint_name)
                    if self.robot:
                        if position < self.robot.qlim[0][index]:
                            position = self.robot.qlim[0][index]
                            self.get_logger().warn(
                                f"joint_position_handler: Joint '{joint_name}' position {position} is out of bounds. Using {self.robot.qlim[0][index]}."
                            )
                        elif position > self.robot.qlim[1][index]:
                            position = self.robot.qlim[1][index]
                            self.get_logger().warn(
                                f"joint_position_handler: Joint '{joint_name}' position {position} is out of bounds. Using {self.robot.qlim[1][index]}."
                            )
                    else:
                        if self.joint_lim[0][index] is not None and position < self.joint_lim[0][index]:
                            position = self.joint_lim[0][index]
                            self.get_logger().warn(
                                f"joint_position_handler: Joint '{joint_name}' position {position} is out of bounds. Using {self.joint_lim[0][index]}."
                            )
                        elif self.joint_lim[1][index] is not None and position > self.joint_lim[1][index]:
                            position = self.joint_lim[1][index]
                            self.get_logger().warn(
                                f"joint_position_handler: Joint '{joint_name}' position {position} is out of bounds. Using {self.joint_lim[1][index]}."
                            )
                    desired_joint_positions = self.current_joint_positions
                    desired_joint_positions[index] = position
                    self.desired = ["joint_positions"] + desired_joint_positions + [duration]
                    self.new = True
                else:
                    self.get_logger().warn(f"joint_position_handler: Invalid number of arguments for joint position. Expected 1, but got {len(args)}.")
            else:
                self.get_logger().warn(f"joint_position_handler: Joint '{joint_name}' not found in the robot model.")
        except Exception as e:
            self.get_logger().fatal(f"joint_position_handler: {e}")

    def cartesian_trajectory_handler(self, *args):
        """Handles incoming OSC messages for cartesian trajectory."""
        if self.robot:
            try:
                if len(args[0]) == 6:
                    points = [[float(j) for j in i] for i in args]
                elif len(args[0]) >= 7:
                    points = [[float(j) for j in i[:6]] for i in args]
                    self.get_logger().warn("cartesian_trajectory_handler: Duration is not supported for cartesian trajectory yet. Ignoring duration.")
                else:
                    self.get_logger().warn("cartesian_trajectory_handler: Invalid number of arguments for cartesian trajectory. Expected 6 ([x, y, z, roll, pitch, yaw]) or 7 ([x, y, z, roll, pitch, yaw, duration]), but got {len(args[0])}.")
                    return
                self.desired = ["cartesian_trajectory"] + points
                self.new = True
            except Exception as e:
                self.get_logger().fatal(f"cartesian_trajectory_handler: {e}")
        else:
            self.get_logger().warn("cartesian_trajectory_handler: No robot model provided. Cannot handle cartesian trajectory.")
            return

    def joint_positions_handler(self, *args):
        """Handles incoming OSC messages for joint positions."""
        try:
            if len(args) == len(self.joint_names):
                desired_joint_positions = [float(i) for i in list(args)]
            elif len(args) == len(self.joint_names) + 1:
                desired_joint_positions = [float(i) for i in list(args)]
            else:
                self.get_logger().warn(f"joint_positions_handler: Invalid number of arguments for joint positions. Expected {len(self.joint_names)} ([q0, q1, q2, ... q{len(self.joint_names)}]) or {len(self.joint_names)+1} ([q0, q1, q2, ... q{len(self.joint_names)}, duration]), but got {len(args)}.")
                return
            # Check if joint positions exceed limits
            if self.robot:
                for i, position in enumerate(desired_joint_positions[:len(self.joint_names)]):  # Exclude duration if present
                    if position < self.robot.qlim[0][i]:
                        desired_joint_positions[i] = self.robot.qlim[0][i]
                        self.get_logger().warn(
                            f"joint_positions_handler: Joint '{self.joint_names[i]}' position {position} is out of bounds. Using {self.robot.qlim[0][i]}."
                        )
                    elif position > self.robot.qlim[1][i]:
                        desired_joint_positions[i] = self.robot.qlim[1][i]
                        self.get_logger().warn(
                            f"joint_positions_handler: Joint '{self.joint_names[i]}' position {position} is out of bounds. Using {self.robot.qlim[1][i]}."
                        )
            else:
                if self.joint_lim:
                    for i, position in enumerate(desired_joint_positions[:len(self.joint_names)]):
                        if self.joint_lim[0][i] is not None:
                            if position < self.joint_lim[0][i]:
                                desired_joint_positions[i] = self.joint_lim[0][i]
                                self.get_logger().warn(
                                    f"joint_positions_handler: Joint '{self.joint_names[i]}' position {position} is out of bounds. Using {self.joint_lim[0][i]}."
                                )
                        elif self.joint_lim[1][i] is not None:
                            if position > self.joint_lim[1][i]:
                                desired_joint_positions[i] = self.joint_lim[1][i]
                                self.get_logger().warn(
                                    f"joint_positions_handler: Joint '{self.joint_names[i]}' position {position} is out of bounds. Using {self.joint_lim[1][i]}."
                                )
            self.desired = ["joint_positions"] + desired_joint_positions
            self.new = True
        except Exception as e:
            self.get_logger().fatal(f"joint_positions_handler: {e}")
            return

    def tcp_coordinates_handler(self, *args):
    # Ensure the desired joint positions are within the specified limits
        if self.robot:
            try:
                if len(args) == 6:
                    x, y, z, r, p, yaw = [float(i) for i in list(args)]
                    duration = None
                elif len(args) >= 7:
                    x, y, z, r, p, yaw, duration, *_ = [float(i) for i in list(args)]
                else:
                    self.get_logger().warn(f"tcp_coordinates_handler: Invalid number of arguments for TCP coordinates. Expected 6 ([x, y, z, roll, pitch, yaw]) or 7 ([x, y, z, roll, pitch, yaw, duration]), but got {len(args)}.")
                    return
                if self.x_limits[0] is not None:
                    x = max(self.x_limits[0], x)
                if self.x_limits[1] is not None:
                    x = min(self.x_limits[1], x)
                if self.y_limits[0] is not None:
                    y = max(self.y_limits[0], y)
                if self.y_limits[1] is not None:
                    y = min(self.y_limits[1], y)
                if self.z_limits[0] is not None:
                    z = max(self.z_limits[0], z)
                if self.z_limits[1] is not None:
                    z = min(self.z_limits[1], z)
                if x != args[0] or y != args[1] or z != args[2]:
                    self.get_logger().warn(
                    f"tcp_coordinates_handler: Desired joint positions adjusted to fit within limits: "
                    f"x={x}, y={y}, z={z} (original: x={args[0]}, y={args[1]}, z={args[2]})"
                    )
                self.desired = ["tcp_coordinates", x, y, z, r, p, yaw, duration]
                self.new = True
            except Exception as e:
                self.get_logger().fatal(f"tcp_coordinates_handler: {e}")
        else:
            self.get_logger().warn("tcp_coordinates_handler: No robot model provided. Cannot handle TCP coordinates.")
            return
            
    def joint_states_callback(self, msg: JointState):
        """Callback function to handle incoming joint states."""
        try:
            msg_time = oscbuildparse.OSCMessage("/time", ',s', [str(time.time())])
            osc_send(msg_time, "osc_client")
            if not(self.joint_names):
                self.joint_names = msg.name
            joint_position_dict = dict(zip(msg.name, msg.position))
            self.current_joint_positions = [float(joint_position_dict[name]) for name in self.joint_names]
            joint_position_dict = dict(zip(msg.name, msg.velocity))
            self.current_joint_velocities = [float(joint_position_dict[name]) for name in self.joint_names]
            if self.robot:
                tcp_position = self.robot.fkine(self.current_joint_positions).t
                tcp_orientation = self.robot.fkine(self.current_joint_positions).rpy()
                msg_tcp = oscbuildparse.OSCMessage("/tcp_coordinates", ',ffffff', [tcp_position[0], tcp_position[1], tcp_position[2], tcp_orientation[0], tcp_orientation[1], tcp_orientation[2]])
                #msg_x = oscbuildparse.OSCMessage(f"/tcp_coordinates/x", ',f', [tcp_position[0]])
                #msg_y = oscbuildparse.OSCMessage(f"/tcp_coordinates/y", ',f', [tcp_position[1]])
                #msg_z = oscbuildparse.OSCMessage(f"/tcp_coordinates/z", ',f', [tcp_position[2]])
                #msg_roll = oscbuildparse.OSCMessage(f"/tcp_coordinates/roll", ',f', [tcp_orientation[0]])
                #msg_pitch = oscbuildparse.OSCMessage(f"/tcp_coordinates/pitch", ',f', [tcp_orientation[1]])
                #msg_yaw = oscbuildparse.OSCMessage(f"/tcp_coordinates/yaw", ',f', [tcp_orientation[2]])
                #bun = oscbuildparse.OSCBundle(oscbuildparse.OSC_IMMEDIATELY, [msg_tcp, msg_x, msg_y, msg_z, msg_roll, msg_pitch, msg_yaw])
                #osc_send(bun, "osc_client") 
                osc_send(msg_tcp, "osc_client")
            msg_position = oscbuildparse.OSCMessage("/joint_state/position", f',{"f"*self.n_joints}', [i for i in msg.position])
            msg_velocity = oscbuildparse.OSCMessage("/joint_state/velocity", f',{"f"*self.n_joints}', [i for i in msg.velocity])
            msg_effort = oscbuildparse.OSCMessage("/joint_state/effort", f',{"f"*self.n_joints}', [i for i in msg.effort])
            msg_name = oscbuildparse.OSCMessage("/joint_state/name", f',{"s"*self.n_joints}', [i for i in msg.name])
            bun = oscbuildparse.OSCBundle(oscbuildparse.OSC_IMMEDIATELY, [msg_name, msg_position, msg_velocity, msg_effort])
            osc_send(bun, "osc_client")
            #for i, name in enumerate(msg.name):
            #    msg_position = oscbuildparse.OSCMessage(f"/joint_state/position/{name}", ',f', [msg.position[i]])
            #    msg_velocity = oscbuildparse.OSCMessage(f"/joint_state/velocity/{name}", ',f', [msg.velocity[i]])
            #    msg_effort = oscbuildparse.OSCMessage(f"/joint_state/effort/{name}", ',f', [msg.effort[i]])
            #    bun = oscbuildparse.OSCBundle(oscbuildparse.OSC_IMMEDIATELY, [msg_position, msg_velocity, msg_effort])
            #    osc_send(bun, "osc_client")
        except Exception as e:
            self.get_logger().fatal(f"joint_states_callback: {e}")

    def send_joint_positions(self):
        self.previous_desired = None
        self.new = False
        try:
            if len(self.desired) == len(self.joint_names) + 2:
                desired_joint_positions = [float(i) for i in self.desired[1:-1]]
                duration = self.desired[-1]
                msg = JointTrajectory()
                msg.joint_names = self.joint_names
                point = JointTrajectoryPoint()
                point.positions = desired_joint_positions
                point.time_from_start.sec = int(duration)
                point.time_from_start.nanosec = int((duration - int(duration)) * 1e9)
                msg.points.append(point)
                self.publisher.publish(msg)
            elif len(self.desired) == len(self.joint_names) + 1:
                desired_joint_positions = [float(i) for i in self.desired[1:]]
                msg = JointTrajectory()
                msg.joint_names = self.joint_names
                point = JointTrajectoryPoint()
                point.positions = desired_joint_positions
                duration = 0
                for p1, p2, max_vel in zip(desired_joint_positions, self.current_joint_positions, self.joint_velocity_limits.values()):
                    duration = max(duration, abs(p1 - p2) / max_vel)
                duration = duration + min(2, 0.2*duration)
                duration /= self.speed_scaling
                if duration == 0:
                    self.get_logger().warn("send_joint_positions: Duration is 0.")
                    return
                point.time_from_start.sec = int(duration)
                point.time_from_start.nanosec = int((duration - int(duration)) * 1e9)
                msg.points.append(point)
                self.publisher.publish(msg)
            else: 
                self.get_logger().warn(f"send_joint_positions: Invalid number of arguments for joint positions. Expected {len(self.joint_names)+1} ([q0, q1, q2, ... q{len(self.joint_names)}, duration]) or {len(self.joint_names)} ([q0, q1, q2, ... q{len(self.joint_names)}]), but got {len(self.desired)}.")
                return
        except Exception as e:
            self.get_logger().fatal(f"send_joint_positions: {e}")
            return

    def trapezoidal_timestamps(self, num_points,total_duration,  flat_ratio = 0.3):
        if num_points == 2:
            return [0, total_duration]
        n = int(num_points*(1-flat_ratio)/2)
        start = np.cos(np.linspace(0, np.pi, n))+2
        end = np.cos(np.linspace(-np.pi, 0, n))+2
        flat = np.ones(num_points-2*n)
        timestamps = np.concatenate((start, flat, end))
        timestamps *= total_duration / timestamps.sum()
        timestamps = np.cumsum(timestamps)
        return timestamps.tolist()

    def send_tcp_coordinates(self):
        """Send the desired TCP coordinates to the robot."""
        try:
            self.new = False
            msg = JointTrajectory()
            msg.joint_names = self.joint_names
            steps_per_m = 100
            if self.previous_desired is None: 
                [x,y,z] = self.robot.fkine(self.current_joint_positions).t
                [roll, pitch, yaw] = self.robot.fkine(self.current_joint_positions).rpy()
            else:
                [x,y,z] = self.previous_desired[1:4]
                [roll, pitch, yaw] = self.previous_desired[4:-1]
            x1, y1, z1, roll1, pitch1, yaw1 = self.desired[1:7]
            self.previous_desired = self.desired
            steps = int(np.linalg.norm(np.array([x1, y1, z1])- self.robot.fkine(self.current_joint_positions).t) * steps_per_m)
            if steps < 2:
                steps = 2
            cart_traj = [sm.SE3([x+(x1-x)/(steps-1)*i, y+(y1-y)/(steps-1)*i, z+(z1-z)/(steps-1)*i]) * sm.SE3.RPY([roll+(roll1-roll)/(steps-1)*i, pitch+(pitch1-pitch)/(steps-1)*i, yaw+(yaw1-yaw)/(steps-1)*i]) for i in range(steps)]
            '''if self.previous_desired:
                [x,y,z] = self.previous_desired[1:4]
                q0 = sm.UnitQuaternion.RPY(self.previous_desired[3], self.previous_desired[4], self.previous_desired[5])        
            else: 
                [x, y, z] = self.robot.fkine(self.current_joint_positions).t
                q0 = sm.UnitQuaternion(self.robot.fkine(self.current_joint_positions).R)

            x1, y1, z1, roll1, pitch1, yaw1 = self.desired[1:7]
            q1 = sm.UnitQuaternion.RPY(roll1, pitch1, yaw1)

            steps = int(np.linalg.norm(np.array([x1, y1, z1]) - self.robot.fkine(self.current_joint_positions).t) * steps_per_m)
            if steps < 2:
                steps = 2
            
            cart_traj = []
            for i in range(steps):
                alpha = i / (steps - 1)

                # Convert to arrays for robust interpolation if needed
                q0_array = q0.vec
                q1_array = q1.vec
                dot = np.dot(q0_array, q1_array)
                dot = np.clip(dot, -1.0, 1.0)

                if abs(dot) > 0.9995:
                    # Linear interpolation + normalization
                    q_interp_array = (1 - alpha) * q0_array + alpha * q1_array
                    q_interp_array = q_interp_array / np.linalg.norm(q_interp_array)
                    q_interp = sm.UnitQuaternion(q_interp_array)
                else:
                    q_interp = q0.interp(q1, alpha)

                # Interpolate translation
                pos_interp = [
                    x + (x1 - x) * alpha,
                    y + (y1 - y) * alpha,
                    z + (z1 - z) * alpha
                ]

                # Compose SE3 transform
                cart_traj.append(sm.SE3(pos_interp) * q_interp.SE3())'''
            if self.desired[-1]:
                timestamps = self.trapezoidal_timestamps(steps, self.desired[-1], 0.8)
                for j in range(steps):
                    sol = self.robot.ik_LM(cart_traj[j], q0=self.current_joint_positions, mask = self.cost_mask, joint_limits = True, method = 'chan') if j == 0 else self.robot.ik_LM(cart_traj[j], q0=prev_sol, mask = self.cost_mask, joint_limits = True, method = 'chan')
                    if sol[1] == 1:
                        fowards = self.robot.fkine_all(sol[0])
                        out_of_bounds = (fowards.t[1:,0] > self.x_limits_workspace[1] if self.x_limits_workspace[1] is not None else False) | (fowards.t[1:,0] < self.x_limits_workspace[0] if self.x_limits_workspace[0] is not None else False) | (fowards.t[1:,1] > self.y_limits_workspace[1] if self.y_limits_workspace[1] is not None else False) | (fowards.t[1:,1] < self.y_limits_workspace[0] if self.y_limits_workspace[0] is not None else False) | (fowards.t[1:,2] > self.z_limits_workspace[1] if self.z_limits_workspace[1] is not None else False) | (fowards.t[1:,2] < self.z_limits_workspace[0] if self.z_limits_workspace[0] is not None else False)
                        if np.any(out_of_bounds):
                            #print(fowards.t)
                            #indices = np.where(out_of_bounds)[0]
                            #print(f"indices: {indices}")
                            self.get_logger().warn("send_tcp_coordinates: One or more links moved out of bounds!")
                            '''
                            for i in indices:
                                try:
                                    print(f"Joint {self.robot.links[i].name} is out of bounds: (x,y,z) = {fowards.t[i]}")
                                except IndexError:
                                    print(f"index {i} is out of bounds, but no corresponding joint found.")
                            self.previous_desired_tcp_position = self.desired_tcp_position
                            '''
                            break
                        duration = timestamps[j]
                        if duration == 0:
                            prev_sol = list(sol[0])
                            continue
                        point = JointTrajectoryPoint()
                        point.positions = list(sol[0])
                        point.time_from_start.sec = int(duration)
                        point.time_from_start.nanosec = int((duration - int(duration)) * 1e9)
                        msg.points.append(point)
                        prev_sol = list(sol[0])
                    else:
                        self.get_logger().warn(f"send_tcp_coordinates: IK could not find a solution for (x,y,z) = {cart_traj[j].t} and (r,p,y) = {cart_traj[j].rpy()}!")
                        prev_sol = self.current_joint_positions
                if len(msg.points) == 0:
                    self.get_logger().warn("send_tcp_coordinates: The resulting trajectory is empty. Either the IK failed or the trajectory is too short.")
                    self.previous_desired = self.desired    
                    return
                msg.header.stamp = self.get_clock().now().to_msg()
                self.publisher.publish(msg)
                self.previous_desired = self.desired
            else:
                prev_duration = 0
                '''
                if self.previous_desired == None: 
                    [x,y,z] = self.robot.fkine(self.current_joint_positions).t
                    [roll, pitch, yaw] = self.robot.fkine(self.current_joint_positions).rpy()
                else:
                    [x,y,z] = self.previous_desired[:3]
                    [roll, pitch, yaw] = self.previous_desired[3:]
                '''
                for j in range(steps):
                    sol = self.robot.ik_LM(cart_traj[j], q0=self.current_joint_positions, mask = self.cost_mask, joint_limits = True, method = 'chan') if j == 0 else self.robot.ik_LM(cart_traj[j], q0=prev_sol, mask = self.cost_mask, joint_limits = True, method = 'chan')
                    if sol[1] == 1:
                        fowards = self.robot.fkine_all(sol[0])
                        out_of_bounds = (fowards.t[1:,0] > self.x_limits_workspace[1] if self.x_limits_workspace[1] is not None else False) | (fowards.t[1:,0] < self.x_limits_workspace[0] if self.x_limits_workspace[0] is not None else False) | (fowards.t[1:,1] > self.y_limits_workspace[1] if self.y_limits_workspace[1] is not None else False) | (fowards.t[1:,1] < self.y_limits_workspace[0] if self.y_limits_workspace[0] is not None else False) | (fowards.t[1:,2] > self.z_limits_workspace[1] if self.z_limits_workspace[1] is not None else False) | (fowards.t[1:,2] < self.z_limits_workspace[0] if self.z_limits_workspace[0] is not None else False)
                        if np.any(out_of_bounds):
                            #print(fowards.t)
                            #indices = np.where(out_of_bounds)[0]
                            #print(f"indices: {indices}")
                            self.get_logger().warn("send_tcp_coordinates: One or more links moved out of bounds!")
                            '''
                            for i in indices:
                                try:
                                    print(f"Joint {self.robot.links[i].name} is out of bounds: (x,y,z) = {fowards.t[i]}")
                                except IndexError:
                                    print(f"index {i} is out of bounds, but no corresponding joint found.")
                            self.previous_desired_tcp_position = self.desired_tcp_position
                            '''
                            break
                        duration = 0
                        prev = self.current_joint_positions if j == 0 else prev_sol
                        for p1, p2, max_vel in zip(sol[0], prev, self.joint_velocity_limits.values()):
                            duration = max(duration, abs(p1 - p2) / max_vel)#, 1/self.hz) # as minimun
                        prev_sol = list(sol[0])
                        if duration == 0:
                            continue
                        point = JointTrajectoryPoint()
                        point.positions = list(sol[0])
                        duration *= 1.2
                        duration /= self.speed_scaling
                        duration += prev_duration
                        prev_duration = duration
                        point.time_from_start.sec = int(duration)
                        point.time_from_start.nanosec = int((duration - int(duration)) * 1e9)
                        msg.points.append(point)
                    else:
                        self.get_logger().warn(f"send_tcp_coordinates: IK could not find a solution for (x,y,z) = {cart_traj[j].t} and (r,p,y) = {cart_traj[j].rpy()}!")
                        prev_sol = self.current_joint_positions
                if len(msg.points) == 0:
                    self.get_logger().warn("send_tcp_coordinates: The resulting trajectory is empty. Either the IK failed or the trajectory is too short.")
                    self.previous_desired = self.desired    
                    return
                msg.header.stamp = self.get_clock().now().to_msg()
                self.publisher.publish(msg)
                self.previous_desired = self.desired
        except Exception as e:
            self.get_logger().fatal(f"send_tcp_coordinates: {e}")

    def send_joint_trajectory(self):
        try:
            self.new = False
            viapoints = np.array([i for i in self.desired[1:]])
            msg = JointTrajectory()
            msg.joint_names = self.joint_names
            traj = rtb.mstraj(viapoints, q0 = self.current_joint_positions ,dt=0.01, tacc=1.5/self.speed_scaling, qdmax=[self.speed_scaling*i for i in self.joint_velocity_limits.values()])
            msg.points = []
            for i in range(len(traj.q)):
                    point = JointTrajectoryPoint()
                    point.positions = list(traj.q[i])
                    point.time_from_start.sec = int(traj.t[i])
                    point.time_from_start.nanosec = int((traj.t[i] - int(traj.t[i])) * 1e9)
                    msg.points.append(point)
            msg.header.stamp = self.get_clock().now().to_msg()
            self.publisher.publish(msg)
            self.previous_desired = None
        except Exception as e:
            print(f'Error in joint_angles_handler: {e}')

    def send_cartesian_trajectory(self):
        try:
            self.new = False
            viapoints = np.array([i[:6] for i in self.desired[1:]])
            msg = JointTrajectory()
            msg.joint_names = self.joint_names
            x,y,z = self.robot.fkine(self.current_joint_positions).t
            r,p,yaw = self.robot.fkine(self.current_joint_positions).rpy()
            q0 = [x, y, z, r, p, yaw]
            traj = rtb.mstraj(viapoints, q0 = q0 ,dt=0.01, tacc=1.5/self.speed_scaling, qdmax=2*self.speed_scaling)
            msg.points = []
            prev_sol = self.current_joint_positions
            n = max(int(len(traj.q)/200), 1)
            for i in range(len(traj.q)):
                T = sm.SE3(traj.q[i][:3]) * sm.SE3.RPY(traj.q[i][3:], order='xyz')
                sol = self.robot.ik_LM(T, q0=prev_sol, mask = self.cost_mask, joint_limits = True)
                if sol[1] == 1:
                    point = JointTrajectoryPoint()
                    point.positions = list(sol[0])
                    point.time_from_start.sec = int(traj.t[i])
                    if traj.t[i] == 0:
                        continue
                    point.time_from_start.nanosec = int((traj.t[i] - int(traj.t[i])) * 1e9)
                    msg.points.append(point)
                    prev_sol = list(sol[0])
                else:
                    self.get_logger().warn(f"send_cartesian_trajectory: IK could not find a solution for (x,y,z) = {traj.q[i][:3]} and (r,p,y) = {traj.q[i][3:]}!")
            msg.header.stamp = self.get_clock().now().to_msg()
            msg.points = msg.points[::n]
            self.publisher.publish(msg)
            self.previous_desired = None
        except Exception as e:
            print(f'Error in joint_angles_handler: {e}')
        self.previous_desired = None

    def update_position(self):
        """Calls the appropriate function to update the robot's position."""
        try:
            if self.desired is None or not(self.new):
                return
            if self.desired[0] == "joint_positions":
                self.new = False
                self.send_joint_positions()
                return
            elif self.desired[0] == "tcp_coordinates":
                self.new = False
                self.send_tcp_coordinates()
                return
            elif self.desired[0] == "joint_trajectory":
                self.new = False
                self.send_joint_trajectory()
                return
            elif self.desired[0] == "cartesian_trajectory":
                self.new = False
                self.send_cartesian_trajectory()
                return
            else:
                self.get_logger().warn(f"update_position: Unknown desired type '{self.desired[0]}'.")
                return
        except Exception as e:
            self.get_logger().fatal(f'update_position: {e}')
                

    def clean_log_string(self, s):
        s = str(s)
        # Remove ANSI escape sequences (e.g., \x1b[31m)
        ansi_escape = re.compile(r'\x1B(?:[@-Z\\-_]|\[[0-?]*[ -/]*[@-~])')
        s = ansi_escape.sub('', s)
        # Replace tabs/newlines with spaces
        s = s.replace('\n', ' ').replace('\r', ' ').replace('\t', ' ').replace("'", ' '). replace('"', ' ').replace('`', ' ').replace('´', ' ').replace('`', ' ').replace('“', ' ').replace('”', ' ').replace('‘', ' ').replace('’', ' ').replace('´', ' ').replace('`', ' ').replace('“', ' ').replace('”', ' ').replace('‘', ' ').replace('’', ' ')
        # Strip leading/trailing whitespace
        s = s.strip()
        # Optionally enforce ASCII only (replace non-ASCII chars with '?')
        s = s.encode('ascii', 'replace').decode('ascii')
        return s

    
    def log_callback(self, msg: Log):
        """Callback function to handle incoming log messages."""
        # Send the log message as an OSC message
        msg_log = oscbuildparse.OSCMessage(f"/log/{self.log_dict.get(msg.level, 'UNKNOWN')}", ',isss', [int(msg.level), str(msg.stamp.sec+msg.stamp.nanosec*1e-9) , str(msg.name), self.clean_log_string(msg.msg)])
        osc_send(msg_log, "osc_log_client")


def main():
    """Main function to get joint names and start the ROS 2 & OSC system."""
    parser = argparse.ArgumentParser(description="An interface between ROS2 and OSC for robotic arm manipulators", usage="ros2 run osc_ros2 interface [-c CONFIG.TOML]")
    parser.add_argument("-c", "--config", type=argparse.FileType("r"), help="A robot TOML config, if the config is provided then the interface will not show the iteractive input")
    args = parser.parse_args()
    rclpy.init()
    if not args.config:
        config, robot = interactive_input()
    else:
        config = toml.load(args.config)
        robot = rtb.ERobot.URDF(config["robot"]["urdf"])
    node = OSC_ROS2_interface(robot, config)
    
    # Run ROS 2 spin, and osc_process will be handled by the timer
    try:
        rclpy.spin(node)
    except KeyboardInterrupt:
        print("")
    finally:
        node.destroy_node()
        osc_terminate()

if __name__ == '__main__':
    main()
